@@ -7,7 +7,7 @@ use Carp ();
 use IO::Handle ();
 use Exporter ();
 
-our $VERSION = '0.013';
+our $VERSION = '0.014';
 
 use constant DEFAULT_REQUEST_BODY_LIMIT => 16777216;
 use constant DEFAULT_REQUEST_BODY_BUFFER => 262144;
@@ -167,7 +167,7 @@ sub _handle_error {
     my ($error_error, $error_errored);
     {
       local $@;
-      eval { $handler->($cgi, $error); 1 } or do { $error_error = $@; $error_errored = 1 };
+      eval { $handler->($cgi, $error, !!$cgi->{headers_rendered}); 1 } or do { $error_error = $@; $error_errored = 1 };
     }
     return unless $cgi->{pid} == $$; # in case of fork in error handler
     if ($error_errored) {
@@ -523,7 +523,7 @@ sub add_response_header {
         my ($key, $val) = @attrs[$i, $i+1];
         my $has_value = $COOKIE_ATTR_VALUE{lc $key};
         if (!defined $has_value) {
-          Carp::carp "Attempted to set unknown cookie attribute '$key' for HTTP response cookie '$name'";
+          Carp::croak "Attempted to set unknown cookie attribute '$key' for HTTP response cookie '$name'";
         } elsif ($has_value) {
           $cookie_str .= "; $key=$val" if defined $val;
         } else {
@@ -549,10 +549,8 @@ sub response_status_code {
   return 200;
 }
 
-sub headers_rendered { $_[0]{headers_rendered} }
-
 {
-  my %RENDER_TYPES = (json => 1, html => 1, xml => 1, text => 1, data => 1, file => 1, handle => 1, redirect => 1);
+  my %RENDER_TYPES = (text => 1, html => 1, xml => 1, json => 1, data => 1, file => 1, handle => 1, redirect => 1);
 
   sub render {
     my ($self, $type, $data) = @_;
@@ -562,10 +560,7 @@ sub headers_rendered { $_[0]{headers_rendered} }
     Carp::croak "Cannot render from an open filehandle with ->render; use ->render_chunk" if $type eq 'handle';
 
     my ($response_body, $response_length, $redirect_url);
-    if ($type eq 'json') {
-      $response_body = $self->_json->encode($data);
-      $response_length = length $response_body;
-    } elsif ($type eq 'html' or $type eq 'xml' or $type eq 'text') {
+    if ($type eq 'text' or $type eq 'html' or $type eq 'xml') {
       my $charset = $self->{response_charset};
       $charset = 'UTF-8' unless defined $charset;
       if (uc $charset eq 'UTF-8' and do { local $@; eval { require Unicode::UTF8; 1 } }) {
@@ -575,6 +570,9 @@ sub headers_rendered { $_[0]{headers_rendered} }
         $response_body = Encode::encode($charset, "$data");
       }
       $response_length = length $response_body;
+    } elsif ($type eq 'json') {
+      $response_body = $self->_json->encode($data);
+      $response_length = length $response_body;
     } elsif ($type eq 'data') {
       $response_body = $data;
       $response_length = length $response_body;
@@ -624,10 +622,7 @@ sub headers_rendered { $_[0]{headers_rendered} }
       $self->{headers_rendered} = 1;
     }
 
-    if ($type eq 'json') {
-      my $response_body = $self->_json->encode($data);
-      $out_fh->printflush($response_body);
-    } elsif ($type eq 'html' or $type eq 'xml' or $type eq 'text') {
+    if ($type eq 'text' or $type eq 'html' or $type eq 'xml') {
       my $charset = $self->{response_charset};
       $charset = 'UTF-8' unless defined $charset;
       my $response_body;
@@ -638,6 +633,9 @@ sub headers_rendered { $_[0]{headers_rendered} }
         $response_body = Encode::encode($charset, "$data");
       }
       $out_fh->printflush($response_body);
+    } elsif ($type eq 'json') {
+      my $response_body = $self->_json->encode($data);
+      $out_fh->printflush($response_body);
     } elsif ($type eq 'data') {
       $out_fh->printflush($data);
     } elsif ($type eq 'file' or $type eq 'handle') {
@@ -692,10 +690,10 @@ sub _response_headers {
     my $charset = $self->{response_charset};
     $charset = 'UTF-8' unless defined $charset;
     $content_type =
-        $type eq 'json' ? 'application/json;charset=UTF-8'
+        $type eq 'text' ? "text/plain;charset=$charset"
       : $type eq 'html' ? "text/html;charset=$charset"
       : $type eq 'xml'  ? "application/xml;charset=$charset"
-      : $type eq 'text' ? "text/plain;charset=$charset"
+      : $type eq 'json' ? 'application/json;charset=UTF-8'
       : 'application/octet-stream'
       unless defined $content_type;
     $headers_str = "Content-Type: $content_type\r\n$headers_str";
